"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const fp_1 = require("lodash/fp");
const types_1 = require("../types");
const utils_1 = require("./utils");
const getEnumValues = (type) => fp_1.map(literal => literal.value, type.types);
const getTurpleNodeDescription = (typeNode, typeArgumentsMap = {}) => checker => ({
    flag: types_1.DescriptionFlag.Turple,
    description: typeNode.elementTypes.map(i => ({
        description: exports.generateNodeDescription(i, typeArgumentsMap)(checker),
    })),
});
const getPropertyDeclarationDescription = (node, typeArgumentsMap = {}) => checker => {
    if (!node || !node.type) {
        return types_1.PropertyType.Null;
    }
    if (ts.isArrayTypeNode(node.type)) {
        const symbol = checker.getSymbolAtLocation(node.name);
        return {
            key: symbol ? symbol.getName() : '',
            flag: types_1.DescriptionFlag.Array,
            description: exports.generateNodeDescription(node.type, typeArgumentsMap)(checker),
        };
    }
    const symbol = checker.getSymbolAtLocation(node.name);
    const type = checker.getTypeAtLocation(node);
    if (!symbol) {
        return types_1.PropertyType.Null;
    }
    return (generateArrayTypeArgumentDescription(node, symbol, typeArgumentsMap)(checker) || {
        key: symbol.getName(),
        flag: ts.isArrayTypeNode(node.type) ? types_1.DescriptionFlag.Array : null,
        description: generatePropertyDescription(symbol.getName(), type, typeArgumentsMap)(checker),
    });
};
const getArrayTypeDescription = (node, typeArgumentsMap = {}) => checker => {
    if (ts.isArrayTypeNode(node.elementType)) {
        return {
            flag: types_1.DescriptionFlag.Array,
            description: exports.generateNodeDescription(node.elementType, typeArgumentsMap)(checker),
        };
    }
    const type = checker.getTypeFromTypeNode(node.elementType);
    if (fp_1.isEmpty(typeArgumentsMap)) {
        return {
            flag: types_1.DescriptionFlag.Array,
            description: generatePropertyDescription(null, type, typeArgumentsMap)(checker),
        };
    }
    return generatePropertyDescription(null, type, typeArgumentsMap)(checker);
};
const getPropertySignatureDescription = (node, typeArgumentsMap = {}) => checker => {
    const symbol = checker.getSymbolAtLocation(node.name);
    if (!symbol || !node.type) {
        return types_1.PropertyType.Null;
    }
    // Return description with type node description
    if (ts.isTypeReferenceNode(node.type)) {
        const type = checker.getTypeAtLocation(node);
        return (generateArrayTypeArgumentDescription(node, symbol, typeArgumentsMap)(checker) || {
            key: symbol.getName(),
            flag: ts.isArrayTypeNode(node.type) ? types_1.DescriptionFlag.Array : null,
            description: generatePropertyDescription(symbol.getName(), type, typeArgumentsMap)(checker),
        });
    }
    // Return description for node if its name is array
    if (ts.isArrayTypeNode(node.type)) {
        const type = checker.getTypeAtLocation(node.type.elementType);
        if (ts.isArrayTypeNode(node.type.elementType)) {
            return {
                key: symbol.getName(),
                flag: types_1.DescriptionFlag.Array,
                description: exports.generateNodeDescription(node.type, typeArgumentsMap)(checker),
            };
        }
        return {
            key: symbol.getName(),
            flag: types_1.DescriptionFlag.Array,
            description: generatePropertyDescription(symbol.getName(), type, typeArgumentsMap)(checker),
        };
    }
    // Return basic type description
    return {
        key: symbol.getName(),
        description: utils_1.getPropertyNameBySyntaxKind(node),
    };
};
const getTypeNodeDescription = node => checker => {
    const type = checker.getTypeFromTypeNode(node);
    const argsMap = utils_1.createTypeArgumentsMap(type)(checker);
    return generatePropertyDescription(null, type, argsMap)(checker);
};
const mergeTypeArgumentsMaps = (argumentsMap = {}, nodeTypeArguments = {}, typeArgumentsMap = {}) => {
    const newArgumentsMap = fp_1.isEmpty(argumentsMap)
        ? fp_1.isEmpty(nodeTypeArguments) ||
            Object.keys(nodeTypeArguments).some(k => typeArgumentsMap[k])
            ? typeArgumentsMap
            : nodeTypeArguments
        : argumentsMap;
    Object.keys(newArgumentsMap).forEach(key => {
        const item = newArgumentsMap[key];
        const symbol = item.type.getSymbol();
        if (!symbol) {
            return;
        }
        const found = typeArgumentsMap[symbol.name] || nodeTypeArguments[symbol.name];
        if (found) {
            newArgumentsMap[key] = {
                isArray: item.isArray || found.isArray,
                type: found.type,
            };
            return;
        }
        utils_1.getTypeArguments(item.type).forEach(i => {
            if (!i.symbol) {
                return;
            }
            const arr = utils_1.getFirstTypeParameter(i);
            const symbolName = utils_1.isArrayType(i)
                ? arr
                    ? arr.symbol && arr.symbol.name
                    : i.symbol.name
                : i.symbol.name;
            if (!symbolName) {
                return;
            }
            newArgumentsMap[symbolName] =
                typeArgumentsMap[symbolName] ||
                    nodeTypeArguments[symbolName] ||
                    newArgumentsMap[symbolName];
        });
    });
    return newArgumentsMap;
};
/**
 * Generate description for property declaration
 * @param node Property declarations
 * @param nodeTypeArguments Property type arguments relations map
 * @param typeArgumentsMap Parent node type arguments relations map
 */
const generateDeclarationDescription = (node, nodeTypeArguments = {}, typeArgumentsMap = {}) => (checker) => {
    if (!node) {
        return types_1.PropertyType.Unknown;
    }
    if (!node.type) {
        const key = node.name.getText();
        return {
            key,
            description: types_1.PropertyType.Unknown,
        };
    }
    if (ts.isTupleTypeNode(node.type)) {
        return getTurpleNodeDescription(node.type)(checker);
    }
    const nodeTypeNode = ts.isArrayTypeNode(node.type)
        ? node.type.elementType
        : node.type;
    const nodeType = checker.getTypeFromTypeNode(nodeTypeNode);
    const argumentsMap = utils_1.createTypeArgumentsMap(nodeType)(checker);
    return exports.generateNodeDescription(node, mergeTypeArgumentsMaps(argumentsMap, nodeTypeArguments, typeArgumentsMap))(checker);
};
/**
 * Generate description for array type argument
 * @param node Node
 * @param symbol Node symbol
 * @param typeArgumentsMap Node type arguments relations map
 */
const generateArrayTypeArgumentDescription = (node, symbol, typeArgumentsMap = {}) => (checker) => {
    const type = checker.getTypeAtLocation(node);
    const isArrayNode = utils_1.isArrayType(type);
    const isTypeArray = type.symbol
        ? typeArgumentsMap[type.symbol.name] &&
            typeArgumentsMap[type.symbol.name].isArray
        : false;
    const isArray = isArrayNode || isTypeArray;
    if (!isArray) {
        return;
    }
    const argumentType = isArrayNode
        ? utils_1.getFirstTypeParameter(type)
        : typeArgumentsMap[type.symbol.name].type;
    return {
        key: symbol.getName(),
        flag: isArray ? types_1.DescriptionFlag.Array : null,
        description: generatePropertyDescription(symbol.getName(), argumentType || typeArgumentsMap[type.symbol.name].type, typeArgumentsMap)(checker),
    };
};
/**
 * Generate property description
 * @param key Property key
 * @param type Type for description generation
 * @param typeArgumentsMap Node type arguments relations
 */
const generatePropertyDescription = (key, type, typeArgumentsMap = {}) => (checker) => {
    if (!type) {
        return types_1.PropertyType.Unknown;
    }
    if (!type.symbol) {
        const typeNode = checker.typeToTypeNode(type);
        return utils_1.getPropertyNameBySyntaxKind(typeNode);
    }
    const nameByWrapper = utils_1.getTypeNameByWrapperFunction(type.symbol.name);
    if (nameByWrapper) {
        return nameByWrapper;
    }
    const arrayTypeParam = utils_1.getFirstTypeParameter(type);
    if (utils_1.isArrayType(type) && arrayTypeParam) {
        return {
            flag: types_1.DescriptionFlag.Array,
            description: generatePropertyDescription(key, arrayTypeParam, typeArgumentsMap)(checker),
        };
    }
    const enumValues = getEnumValues(type);
    if (!fp_1.isEmpty(enumValues)) {
        return {
            flag: types_1.DescriptionFlag.Enum,
            possibleValues: enumValues,
            description: [],
        };
    }
    const declarations = fp_1.flatMap(p => p.declarations, checker.getPropertiesOfType(type));
    if (!declarations.length) {
        const typeArg = typeArgumentsMap[type.symbol.name];
        const isArray = typeArg && typeArg.isArray;
        if (isArray) {
            return {
                flag: types_1.DescriptionFlag.Array,
                description: generatePropertyDescription(key, typeArg && typeArg.type, typeArgumentsMap)(checker),
            };
        }
        return generatePropertyDescription(key, typeArg && typeArg.type, typeArgumentsMap)(checker);
    }
    const nodeTypeArguments = utils_1.createTypeArgumentsMap(type)(checker);
    return fp_1.flatMap(node => generateDeclarationDescription(node, nodeTypeArguments, typeArgumentsMap)(checker), declarations);
};
/**
 * Generates description for node type
 * @param node Target node
 * @param typeArgumentsMap Node type arguments relations
 */
exports.generateNodeDescription = (node, typeArgumentsMap = {}) => checker => {
    // Return description for array type node
    if (ts.isArrayTypeNode(node)) {
        return getArrayTypeDescription(node, typeArgumentsMap)(checker);
    }
    // Return description for turple type node
    if (ts.isTupleTypeNode(node)) {
        return getTurpleNodeDescription(node)(checker);
    }
    // Return description for type node
    if (ts.isTypeNode(node)) {
        return getTypeNodeDescription(node)(checker);
    }
    // Return property declaration and array of properties declarations description
    if (ts.isPropertyDeclaration(node)) {
        return getPropertyDeclarationDescription(node, typeArgumentsMap)(checker);
    }
    // Then we need property signatures
    if (!ts.isPropertySignature(node)) {
        return types_1.PropertyType.Null;
    }
    return getPropertySignatureDescription(node, typeArgumentsMap)(checker);
};
//# sourceMappingURL=description-generate.js.map