"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const fp_1 = require("lodash/fp");
const types_1 = require("../types");
const TARGET_CALLERS = ['create', 'createMany', 'build'];
const TARGET_CLASS_NAME = 'Randomizer';
/**
 * Create type arguments map where key is type argument symbol name, value is concrete argument type
 * @param type Type
 */
exports.createTypeArgumentsMap = (type) => (checker) => {
    const argumentsMap = {};
    const symbol = type.getSymbol();
    if (!symbol) {
        return argumentsMap;
    }
    const typeParameters = fp_1.flatMap(d => ts.isInterfaceDeclaration(d) || ts.isClassDeclaration(d)
        ? d.typeParameters
        : [], symbol.declarations);
    if (exports.isArrayType(type)) {
        const arrayTypeParam = exports.getFirstTypeParameter(type);
        if (arrayTypeParam) {
            return exports.createTypeArgumentsMap(arrayTypeParam)(checker);
        }
    }
    return exports.getTypeArguments(type).reduce((acc, item, index) => {
        const typeParameterItem = typeParameters[index];
        if (!typeParameterItem) {
            return acc;
        }
        const itemSymbol = checker.getTypeAtLocation(typeParameterItem).getSymbol();
        const isArray = exports.isArrayType(item);
        const type = isArray ? exports.getFirstTypeParameter(item) : item;
        if (!type) {
            return acc;
        }
        if (itemSymbol) {
            acc[itemSymbol.name] = {
                type,
                isArray,
            };
        }
        return acc;
    }, argumentsMap);
};
exports.getTypeNameByWrapperFunction = (name) => {
    switch (name) {
        case 'Number':
            return types_1.PropertyType.Number;
        case 'String':
            return types_1.PropertyType.String;
        case 'Boolean':
            return types_1.PropertyType.Boolean;
        case 'Date':
            return types_1.PropertyType.Date;
        case 'Function':
            return types_1.PropertyType.Function;
        case 'Object':
            return types_1.PropertyType.Object;
        default:
            return null;
    }
};
exports.getPropertyNameBySyntaxKind = (propertySignature) => {
    if (!propertySignature) {
        return types_1.PropertyType.Unknown;
    }
    let kind = propertySignature.kind;
    if (ts.isPropertyDeclaration(propertySignature) ||
        ts.isPropertySignature(propertySignature)) {
        kind = propertySignature.type && propertySignature.type.kind;
    }
    switch (kind) {
        case ts.SyntaxKind.StringKeyword:
            return types_1.PropertyType.String;
        case ts.SyntaxKind.NumberKeyword:
            return types_1.PropertyType.Number;
        case ts.SyntaxKind.BooleanKeyword:
            return types_1.PropertyType.Boolean;
        case ts.SyntaxKind.FunctionKeyword:
            return types_1.PropertyType.Function;
        case ts.SyntaxKind.ObjectKeyword:
            return types_1.PropertyType.Object;
        case ts.SyntaxKind.NullKeyword:
            return types_1.PropertyType.Null;
        case ts.SyntaxKind.UndefinedKeyword:
            return types_1.PropertyType.Undefined;
        case ts.SyntaxKind.ArrayType:
            return ts.isPropertyDeclaration(propertySignature) ||
                ts.isPropertySignature(propertySignature)
                ? exports.getPropertyNameBySyntaxKind(propertySignature.type.elementType)
                : exports.getPropertyNameBySyntaxKind(propertySignature.elementType);
        default:
            return types_1.PropertyType.Unknown;
    }
};
exports.getTypeArguments = (type) => type.typeArguments || [];
exports.getFirstTypeParameter = (item) => fp_1.first(exports.getTypeArguments(item));
exports.isArrayType = (type) => type.symbol && type.symbol.name === 'Array';
exports.isTargetExpression = (target) => ts.isPropertyAccessExpression(target.expression) &&
    ts.isIdentifier(target.expression.expression) &&
    TARGET_CALLERS.includes(target.expression.name.text) &&
    target.expression.expression.text === TARGET_CLASS_NAME;
//# sourceMappingURL=utils.js.map